//#include "stdafx.h"

#include "chisqrdistri.h"


#include "adapIntgrtn.h"

using namespace adapInt;

namespace ChiSqrD {

	int df;
	double cinv_chi2;

	double Chi_kernel( double x ) {
		double v = Math::Pow( x, df/2.0 - 1.0 ) * Math::Exp(-x/2.0);
		return v;
	}

	double gamma( double r ) {
		if ( r > 1.25 ) {
			return (r - 1.0) * gamma(r - 1.0);
		} else {
			if (r > 0.75) {
				return 1.0;
			} else {
				return Math::Sqrt(Math::PI);
			};
		};
	}

	double Normal_Dist( double x ) {
		return Math::Exp( -x * x / 2.0 );
	}

	double chi2_density( double x ){
		double v = Math::Exp(-x / 2.0);
		if (v > 0.0) v *= Math::Pow( x, df/2.0 - 1 );
		return v * cinv_chi2;
	}

	double Cum_Chi_distribution( double chi ) {
		if ( chi <= 0.0 ) {
			return 0.0;
		} else {
			if ( df <= 1 ) {
				AdaptiveGL gl;
				return gl.integral( -Math::Sqrt(chi), Math::Sqrt(chi), Normal_Dist ) / Math::Sqrt(2.0 * Math::PI);
			} else {
				AdaptiveGL gl;
				double v = gl.integral( 0.0, chi, chi2_density );
				return v;
			};
		};
	}


	//     Bisection@ɂ鍪̌vZ    

	void Bisection( double (*f)(double),
					double s,					 //   f֐̂Ƃ s > 0.0
					double c,					 //   ɑ΂֐l
					double L_b, double U_b,		 //   ̑݋Ԃ̉Ə
					double & Root,				 //     f(Root) = c
					double acc ){
		if ( (f(L_b) - c) * (f(U_b) - c) > 0.0 ){
			throw  gcnew Exception("Ɖ̑gݍ킹sK؂łB");
		}
		do {
			double m = 0.5 * (L_b + U_b);
			double v = f(m);
			if ( s * (v - c) > 0.0 ) {
				U_b = m;
			} else {
				L_b = m;
			}
			System::Diagnostics::Debug::WriteLine("L_b = " + L_b.ToString() + "   m = " + m.ToString() + "   U_b = " + U_b.ToString());
		} while ( (Math::Abs(U_b - L_b) >= acc * Math::Abs(U_b))
					&&
				  ( (Math::Abs(U_b) + Math::Abs(L_b)) >= 1.0e-14)
				);
		Root = 0.5 * (L_b + U_b);
	}





//     Newton@ɂ鍪̌vZ   

	void Newton( double & x,				//  ̏lA߂l͍ f(x)=c
				 double c,					//  ֐l
				 double (*f)( double ),		//  ֐
				 double (*df)( double ),	//  ֐
				 double acc ){
		 int NStep = 0;
		 int ck = 0;
		 double x0;
		 do {
			 NStep++;
			 if (NStep > 20) {			//  20ȏJԂ̂Ƃ
				 ck = 1;
				 break;
			 };
			 x0 = x;
			 x = x0 + (c - f(x0)) / df(x0);
			 if (x <= 0.0) x = 0.5 * x0;
			 if ( x > x0 * 2.0) x = x0 * 2.0;
		 } while ( (Math::Abs(x - x0) >= (acc * (Math::Abs(x) + Math::Abs(x0))))
					 &&
					 ( (Math::Abs(x) + Math::Abs(x0)) >= 1.0e-14 ) 
			      );

		 if ((ck == 0) && (x > 1.0e-16)) return;

         //   Bisection@ŌvZ

		 double L_b = 1.0;
		 do {
				L_b *= 0.5;
		 } while ( f(L_b) >= c );

		 double U_b = 1.0;
		 do {
				U_b += 1.0;
		 } while ( f(U_b) <= c );

		 Bisection( f, 1.0, c, L_b, U_b, x, acc );
		 return;
	}



	double calc_chi2_for_p( double p ){
		double x2 = 1.0;
		Newton( x2, p, Cum_Chi_distribution, chi2_density, 1.0e-12 );

		return x2;
	}


}

